import aiger
import hypothesis.strategies as st
from aiger import hypothesis as aigh
from aiger_bv import atom
from hypothesis import given, settings

from aiger_bdd import to_bdd, from_bdd, count, bdd_to_nx


@given(aigh.Circuits)
def test_bdd_transform(circ):
    circ = circ.unroll(3)
    out = list(circ.outputs)[0]
    f, manager, relabel = to_bdd(circ, output=out)
    expr = from_bdd(f, manager=manager)

    circ2 = expr.aig['i', relabel.inv]['o', {expr.output: out}]
    assert not circ2.latches
    assert circ2.inputs <= circ.inputs
    assert out in circ2.outputs

    f2, _, relabel = to_bdd(
        expr.aig, manager=manager, output=expr.output, renamer=lambda *x: x[1]
    )
    assert f == f2


def test_bdd_transform_smoke():
    to_bdd(atom(3, 'x', signed=False) < 4)


@settings(max_examples=4, deadline=None)
@given(st.integers(0, 7))
def test_count_le(i):
    expr = atom(4, 'x', signed=False) < atom(4, i, signed=False)
    assert count(expr) == i
    assert count(expr, fraction=True) == i / (2**4)


def test_set_levels():
    a1, c1, a2, c2 = aiger.atoms('a1', 'c1', 'a2', 'c2')
    expr = (a1 == c1) & (a2 == c2)
    levels = {'c1': 0, 'a1': 1, 'c2': 2, 'a2': 3}

    bexpr, manager, relabels = to_bdd(
        expr, renamer=lambda _, x: x, levels=levels
    )

    assert all(k == v for k, v in relabels.items())
    assert bexpr.low.negated
    assert not bexpr.high.negated


def test_bdd_to_nx():
    x, y, z, w = aiger.atoms('x', 'y', 'z', 'w')
    bexpr, *_ = to_bdd(x ^ y ^ z ^ w)
    dag = bdd_to_nx(bexpr)
    assert len(dag) > len(bexpr)
    assert len(bexpr) == 5
    assert len(dag) == 9



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['aiger_bdd']

package_data = \
{'': ['*']}

install_requires = \
['attrs>=22.0.0,<23.0.0',
 'bidict>=0.22.0,<0.23.0',
 'dd>=0.5.4,<0.6.0',
 'funcy>=1.12,<2.0',
 'py-aiger>=6.0.0,<7.0.0']

setup_kwargs = {
    'name': 'py-aiger-bdd',
    'version': '3.1.1',
    'description': 'Aiger to BDD bridge.',
    'long_description': "# py-aiger-bdd\n[![Build Status](https://cloud.drone.io/api/badges/mvcisback/py-aiger-bdd/status.svg)](https://cloud.drone.io/mvcisback/py-aiger-bdd)\n[![codecov](https://codecov.io/gh/mvcisback/py-aiger-bdd/branch/master/graph/badge.svg)](https://codecov.io/gh/mvcisback/py-aiger-bdd)\n\n[![PyPI version](https://badge.fury.io/py/py-aiger-bdd.svg)](https://badge.fury.io/py/py-aiger-bdd)\n[![License: MIT](https://img.shields.io/badge/License-MIT-yellow.svg)](https://opensource.org/licenses/MIT)\n\n# Installation\n\n`$ pip install py-aiger-bdd`\n\nFor developers, note that this project uses the\n[poetry](https://poetry.eustace.io/) python package/dependency\nmanagement tool. Please familarize yourself with it and then\nrun:\n\n`$ poetry install`\n\n# Usage\n\nThis tutorial assumes familiarity with [py-aiger](https://github.com/mvcisback/py-aiger) and [py-aiger-bv](https://github.com/mvcisback/py-aiger-bv).\n\n```python\nimport aiger_bv as BV\nfrom aiger_bdd import to_bdd, from_bdd, count\n\nx = BV.atom(3, 'x', signed=False) \n\nexpr = x < 5  # Could be an AIG or AIGBV or BoolExpr.\nbdd, manager, input2var = to_bdd(expr)  # Convert circuit encoded by expr into a BDD.\nexpr2 = from_bdd(bdd)  # Creates an Aiger Expression from a BDD.\n\nassert count(expr, fraction=True) == 5/8\nassert count(expr, fraction=False) == 5\n```\n",
    'author': 'Marcell Vazquez-Chanlatte',
    'author_email': 'mvc@linux.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/mvcisback/py-aiger-bdd',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import aiger
import funcy as fn
from bidict import bidict
from fractions import Fraction

try:
    from dd.cudd import BDD, Function
except ImportError:
    from dd.autoref import BDD, Function


def to_bdd(circ_or_expr, output=None, manager=None, renamer=None, levels=None):
    if renamer is None:
        _count = 0

        def renamer(*_):
            nonlocal _count
            _count += 1
            return f"x{_count}"

    if not isinstance(circ_or_expr, aiger.BoolExpr):
        circ = aiger.to_aig(circ_or_expr, allow_lazy=True)
        assert len(circ.latches) == 0

        if output is None:
            assert len(circ.outputs) == 1
            output = fn.first(circ.outputs)

        expr = aiger.BoolExpr(circ)
    else:
        expr = circ_or_expr

    manager = BDD() if manager is None else manager
    input_refs_to_var = {
        ref: renamer(i, ref) for i, ref in enumerate(expr.inputs)
    }

    manager.declare(*input_refs_to_var.values())
    if levels is not None:
        assert set(manager.vars) <= set(levels)
        levels = fn.project(levels, set(manager.vars))
        levels = fn.walk_keys(input_refs_to_var.get, levels)

        manager.reorder(levels)
        manager.configure(reordering=False)

    def lift(obj):
        if isinstance(obj, bool):
            return manager.true if obj else manager.false
        return obj

    inputs = {i: manager.var(input_refs_to_var[i]) for i in expr.inputs}
    out = expr(inputs, lift=lift)
    return out, out.bdd, bidict(input_refs_to_var)


def from_bdd(node, manager=None):
    if manager is None:
        manager = node.bdd

    name = node.var
    if name is None:  # Must be a leaf True xor False node.
        return aiger.atom(node == manager.true)

    test = aiger.atom(name)
    low, high = [from_bdd(c, manager) for c in (node.low, node.high)]
    expr = aiger.ite(test, high, low)
    return ~expr if node.negated else expr


def count(circ_or_expr, fraction=False, output=None):
    f, *_ = to_bdd(circ_or_expr, output)
    if not isinstance(circ_or_expr, aiger.AIG):
        circ_or_expr = circ_or_expr.aig

    n_inputs = len(circ_or_expr.inputs)
    num_models = int(f.count(n_inputs))
    return Fraction(num_models, (2**n_inputs)) if fraction else num_models


def bdd_to_nx(bexpr: Function) -> Function:
    import networkx as nx
    # DFS to translate edge-compelemented BDD to networkx graph.
    dag = nx.DiGraph()

    stack, visited = [(bexpr, False, int(bexpr))], set()
    while stack:
        bexpr, parity, ref = stack.pop()

        if ref in visited:
            continue

        visited.add(ref)
        if bexpr in (bexpr.bdd.true, bexpr.bdd.false):
            label = (bexpr == bexpr.bdd.true) ^ parity
            dag.add_node(ref, label=label, level=len(bexpr.bdd.vars))
            continue

        dag.add_node(ref, label=bexpr.var, level=bexpr.level)

        parity = bexpr.negated ^ parity
        for lbl, bexpr2 in [(0, bexpr.low), (1, bexpr.high)]:
            ref2 = int(bexpr2 if parity else ~bexpr2)
            dag.add_edge(ref, ref2, label=lbl)
            stack.append((bexpr2, parity, ref2))
    return dag
